#include <winresrc.h>
#include "config.h"
#define TO_STR_HELPER(x) #x
#define TO_STR(x) TO_STR_HELPER(x)

VS_VERSION_INFO VERSIONINFO
	FILEVERSION OPENSC_VERSION_MAJOR,OPENSC_VERSION_MINOR,OPENSC_VERSION_FIX,OPENSC_VERSION_REVISION
	PRODUCTVERSION OPENSC_VERSION_MAJOR,OPENSC_VERSION_MINOR,OPENSC_VERSION_FIX,OPENSC_VERSION_REVISION
	FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
	FILEFLAGS 0x21L
#else
	FILEFLAGS 0x20L
#endif
	FILEOS 0x40004L
	FILETYPE 0x2L
	FILESUBTYPE 0x0L
BEGIN
	BLOCK "StringFileInfo"
	BEGIN
		BLOCK "040904b0"
		BEGIN
			VALUE "Comments", OPENSC_VS_FF_COMMENTS
			VALUE "CompanyName", OPENSC_VS_FF_COMPANY_NAME
			VALUE "FileVersion", TO_STR(OPENSC_VERSION_MAJOR.OPENSC_VERSION_MINOR.OPENSC_VERSION_FIX.OPENSC_VERSION_REVISION)
			VALUE "InternalName", PACKAGE_NAME
			VALUE "LegalCopyright", OPENSC_VS_FF_LEGAL_COPYRIGHT
			VALUE "LegalTrademarks", ""
			VALUE "PrivateBuild", ""
			VALUE "ProductName", OPENSC_VS_FF_PRODUCT_NAME
			VALUE "ProductVersion", TO_STR(OPENSC_VERSION_MAJOR.OPENSC_VERSION_MINOR.OPENSC_VERSION_FIX.OPENSC_VERSION_REVISION)
			VALUE "SpecialBuild", ""
			VALUE "FileDescription", "OpenSC PKCS#11 module"
		END
	END
	BLOCK "VarFileInfo"
	BEGIN
		VALUE "Translation", 0x409, 1200
	END
END

