/*
 * card-dtrust.h: Support for (CardOS based) D-Trust Signature Cards
 *
 * Copyright (C) 2024 Mario Haustein <mario.haustein@hrz.tu-chemnitz.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef _CARD_DTRUST_H
#define _CARD_DTRUST_H

#define DTRUST4_PIN_ID_PIN_CH	 0x03
#define DTRUST4_PIN_ID_PUK_CH	 0x04
#define DTRUST4_PIN_ID_PIN_T	 0x0b
#define DTRUST4_PIN_ID_QES	 0x87
#define DTRUST5_PIN_ID_PIN_T	 0x0b
#define DTRUST5_PIN_ID_PIN_T_AUT 0x0c
#define DTRUST5_PIN_ID_QES	 0x87
#define DTRUST5_PIN_ID_AUT	 0x91

#endif
